@@ -20,10 +20,13 @@ import bzrlib.errors as errors
 from bzrlib.inventory import InventoryEntry
 from bzrlib.trace import mutter, note, warning
 from bzrlib.errors import NotBranchError
-from bzrlib.branch import Branch
-from bzrlib.osutils import quotefn
+from bzrlib.workingtree import is_control_file
+import bzrlib.osutils
+from bzrlib.workingtree import WorkingTree
 
 def glob_expand_for_win32(file_list):
+    if not file_list:
+        return
     import glob
     expanded_file_list = []
     for possible_glob in file_list:
@@ -38,26 +41,37 @@ def glob_expand_for_win32(file_list):
     return expanded_file_list
 
 
-def add_reporter_null(path, kind, entry):
-    """Absorb add reports and do nothing."""
-    pass
-
-def add_reporter_print(path, kind, entry):
-    """Print a line to stdout for each file that's added."""
-    print "added", quotefn(path)
-    
 def _prepare_file_list(file_list):
     """Prepare a file list for use by smart_add_*."""
     import sys
     if sys.platform == 'win32':
         file_list = glob_expand_for_win32(file_list)
     if not file_list:
-        file_list = ['.']
+        file_list = [u'.']
     file_list = list(file_list)
     return file_list
 
 
-def smart_add(file_list, recurse=True, reporter=add_reporter_null):
+def add_action_null(inv, path, kind):
+    """Absorb add actions and do nothing."""
+    pass
+
+def add_action_print(inv, path, kind):
+    """Print a line to stdout for each file that would be added."""
+    print "added", bzrlib.osutils.quotefn(path)
+
+def add_action_add(inv, path, kind):
+    """Add each file to the given inventory. Produce no output."""
+    entry = inv.add_path(path, kind=kind)
+    mutter("added %r kind %r file_id={%s}" % (path, kind, entry.file_id))
+
+def add_action_add_and_print(inv, path, kind):
+    """Add each file to the given inventory, and print a line to stdout."""
+    add_action_add(inv, path, kind)
+    add_action_print(inv, path, kind)
+
+
+def smart_add(file_list, recurse=True, action=add_action_add):
     """Add files to version, optionally recursing into directories.
 
     This is designed more towards DWIM for humans than API simplicity.
@@ -66,37 +80,39 @@ def smart_add(file_list, recurse=True, reporter=add_reporter_null):
     Returns the number of files added.
     """
     file_list = _prepare_file_list(file_list)
-    b = Branch.open_containing(file_list[0])
-    return smart_add_branch(b, file_list, recurse, reporter)
+    tree = WorkingTree.open_containing(file_list[0])[0]
+    return smart_add_tree(tree, file_list, recurse, action)
 
-        
-def smart_add_branch(branch, file_list, recurse=True, reporter=add_reporter_null):
+def smart_add_tree(tree, file_list, recurse=True, action=add_action_add):
     """Add files to version, optionally recursing into directories.
 
     This is designed more towards DWIM for humans than API simplicity.
     For the specific behaviour see the help for cmd_add().
 
-    This yields a sequence of (path, kind, file_id) for added files.
+    This calls reporter with each (path, kind, file_id) of added files.
 
     Returns the number of files added.
     """
-    import os
+    import os, errno
     from bzrlib.errors import BadFileKindError, ForbiddenFileError
-    import bzrlib.branch
-
     assert isinstance(recurse, bool)
 
     file_list = _prepare_file_list(file_list)
     user_list = file_list[:]
-    inv = branch.read_working_inventory()
-    tree = branch.working_tree()
-    count = 0
+    inv = tree.read_working_inventory()
+    added = []
+    ignored = {}
 
     for f in file_list:
-        rf = branch.relpath(f)
-        af = branch.abspath(rf)
+        rf = tree.relpath(f)
+        af = tree.abspath(rf)
 
-        kind = bzrlib.osutils.file_kind(af)
+        try:
+            kind = bzrlib.osutils.file_kind(af)
+        except OSError, e:
+            if hasattr(e, 'errno') and e.errno == errno.ENOENT:
+                raise errors.NoSuchFile(rf)
+            raise
 
         if not InventoryEntry.versionable_kind(kind):
             if f in user_list:
@@ -105,16 +121,16 @@ def smart_add_branch(branch, file_list, recurse=True, reporter=add_reporter_null
                 warning("skipping %s (can't add file of kind '%s')", f, kind)
                 continue
 
-        mutter("smart add of %r, abs=%r" % (f, af))
+        mutter("smart add of %r, abs=%r", f, af)
         
-        if bzrlib.branch.is_control_file(af):
+        if tree.is_control_filename(af):
             raise ForbiddenFileError('cannot add control file %s' % f)
             
         versioned = (inv.path2id(rf) != None)
 
         if kind == 'directory':
             try:
-                sub_branch = Branch.open(af)
+                sub_branch = WorkingTree.open(af)
                 sub_tree = True
             except NotBranchError:
                 sub_tree = False
@@ -124,48 +140,50 @@ def smart_add_branch(branch, file_list, recurse=True, reporter=add_reporter_null
             sub_tree = False
 
         if rf == '':
-            mutter("branch root doesn't need to be added")
+            mutter("tree root doesn't need to be added")
             sub_tree = False
         elif versioned:
-            mutter("%r is already versioned" % f)
+            mutter("%r is already versioned", f)
         elif sub_tree:
-            mutter("%r is a bzr tree" %f)
+            mutter("%r is a bzr tree", f)
         else:
-            count += __add_one(branch, inv, rf, kind, reporter)
+            added.extend(__add_one(tree, inv, rf, kind, action))
 
         if kind == 'directory' and recurse and not sub_tree:
             for subf in os.listdir(af):
-                subp = os.path.join(rf, subf)
-                if subf == bzrlib.BZRDIR:
-                    mutter("skip control directory %r" % subp)
-                elif tree.is_ignored(subp):
-                    mutter("skip ignored sub-file %r" % subp)
+                subp = bzrlib.osutils.pathjoin(rf, subf)
+                if tree.is_control_filename(subp):
+                    mutter("skip control directory %r", subp)
                 else:
-                    mutter("queue to add sub-file %r" % subp)
-                    file_list.append(branch.abspath(subp))
+                    ignore_glob = tree.is_ignored(subp)
+                    if ignore_glob is not None:
+                        mutter("skip ignored sub-file %r", subp)
+                        if ignore_glob not in ignored:
+                            ignored[ignore_glob] = []
+                        ignored[ignore_glob].append(subp)
+                    else:
+                        mutter("queue to add sub-file %r", subp)
+                        file_list.append(tree.abspath(subp))
 
 
-    mutter('added %d entries', count)
+    mutter('added %d entries', len(added))
     
-    if count > 0:
-        branch._write_inventory(inv)
+    if len(added) > 0:
+        tree._write_inventory(inv)
 
-    return count
+    return added, ignored
 
-def __add_one(branch, inv, path, kind, reporter):
+def __add_one(tree, inv, path, kind, action):
     """Add a file or directory, automatically add unversioned parents."""
 
     # Nothing to do if path is already versioned.
-    # This is safe from infinite recursion because the branch root is
+    # This is safe from infinite recursion because the tree root is
     # always versioned.
     if inv.path2id(path) != None:
-        return 0
+        return []
 
     # add parent
-    count = __add_one(branch, inv, dirname(path), 'directory', reporter)
+    added = __add_one(tree, inv, dirname(path), 'directory', action)
+    action(inv, path, kind)
 
-    entry = inv.add_path(path, kind=kind)
-    mutter("added %r kind %r file_id={%s}" % (path, kind, entry.file_id))
-    reporter(path, kind, entry)
-
-    return count + 1
+    return added + [path]
