@@ -17,7 +17,8 @@
 
 
 from warnings import warn
-from osutils import rename
+from bzrlib.osutils import rename
+import errno
 
 class AtomicFile(object):
     """A file that does an atomic-rename to move into place.
@@ -30,7 +31,7 @@ class AtomicFile(object):
     An encoding can be specified; otherwise the default is ascii.
     """
 
-    def __init__(self, filename, mode='wb', encoding=None):
+    def __init__(self, filename, mode='wb', encoding=None, new_mode=None):
         if mode != 'wb' and mode != 'wt':
             raise ValueError("invalid AtomicFile mode %r" % mode)
 
@@ -47,6 +48,7 @@ class AtomicFile(object):
 
         self.write = self.f.write
         self.closed = False
+        self._new_mode = new_mode
 
 
     def __repr__(self):
@@ -65,6 +67,15 @@ class AtomicFile(object):
         self.f.close()
         self.f = None
         
+        try:
+            if self._new_mode is None:
+                self._new_mode = os.lstat(self.realfilename).st_mode
+        except OSError, e:
+            if e.errno != errno.ENOENT:
+                raise
+        else:
+            os.chmod(self.tmpfilename, self._new_mode)
+
         rename(self.tmpfilename, self.realfilename)
 
 
