@@ -56,6 +56,12 @@
 # merges from, then it should still be reported as newly added
 # relative to the basis revision.
 
+# TODO: Do checks that the tree can be committed *before* running the 
+# editor; this should include checks for a pointless commit and for 
+# unknown or missing files.
+
+# TODO: If commit fails, leave the message in a file somewhere.
+
 
 import os
 import re
@@ -66,25 +72,29 @@ import pdb
 from binascii import hexlify
 from cStringIO import StringIO
 
-from bzrlib.osutils import (local_time_offset, username,
-                            rand_bytes, compact_date, user_email,
+from bzrlib.atomicfile import AtomicFile
+from bzrlib.osutils import (local_time_offset,
+                            rand_bytes, compact_date,
                             kind_marker, is_inside_any, quotefn,
-                            sha_string, sha_strings, sha_file, isdir, isfile,
+                            sha_file, isdir, isfile,
                             split_lines)
-from bzrlib.branch import gen_file_id
+import bzrlib.config
+import bzrlib.errors as errors
 from bzrlib.errors import (BzrError, PointlessCommit,
                            HistoryMissing,
-                           ConflictsInTree
+                           ConflictsInTree,
+                           StrictCommitFailed
                            )
 from bzrlib.revision import Revision
+from bzrlib.testament import Testament
 from bzrlib.trace import mutter, note, warning
 from bzrlib.xml5 import serializer_v5
-from bzrlib.inventory import Inventory, ROOT_ID
-from bzrlib.weave import Weave
-from bzrlib.weavefile import read_weave, write_weave_v5
-from bzrlib.atomicfile import AtomicFile
+from bzrlib.inventory import Inventory, ROOT_ID, InventoryEntry
+from bzrlib.symbol_versioning import *
+from bzrlib.workingtree import WorkingTree
 
 
+@deprecated_function(zero_seven)
 def commit(*args, **kwargs):
     """Commit a new revision to a branch.
 
@@ -114,13 +124,24 @@ class NullCommitReporter(object):
     def missing(self, path):
         pass
 
+    def renamed(self, change, old_path, new_path):
+        pass
+
+
 class ReportCommitToLog(NullCommitReporter):
 
+    # this may be more useful if 'note' was replaced by an overridable
+    # method on self, which would allow more trivial subclassing.
+    # alternative, a callable could be passed in, allowing really trivial
+    # reuse for some uis. RBC 20060511
+
     def snapshot_change(self, change, path):
+        if change == 'unchanged':
+            return
         note("%s %s", change, path)
 
     def completed(self, revno, rev_id):
-        note('committed r%d {%s}', revno, rev_id)
+        note('Committed revision %d.', revno)
     
     def deleted(self, file_id):
         note('deleted %s', file_id)
@@ -131,6 +152,10 @@ class ReportCommitToLog(NullCommitReporter):
     def missing(self, path):
         note('missing %s', path)
 
+    def renamed(self, change, old_path, new_path):
+        note('%s %s => %s', change, old_path, new_path)
+
+
 class Commit(object):
     """Task of committing a new revision.
 
@@ -144,24 +169,39 @@ class Commit(object):
             working inventory.
     """
     def __init__(self,
-                 reporter=None):
+                 reporter=None,
+                 config=None):
         if reporter is not None:
             self.reporter = reporter
         else:
             self.reporter = NullCommitReporter()
-
+        if config is not None:
+            self.config = config
+        else:
+            self.config = None
         
     def commit(self,
-               branch, message,
+               branch=DEPRECATED_PARAMETER, message=None,
                timestamp=None,
                timezone=None,
                committer=None,
                specific_files=None,
                rev_id=None,
                allow_pointless=True,
-               verbose=False):
+               strict=False,
+               verbose=False,
+               revprops=None,
+               working_tree=None,
+               local=False,
+               reporter=None,
+               config=None):
         """Commit working copy as a new revision.
 
+        branch -- the deprecated branch to commit to. New callers should pass in 
+                  working_tree instead
+
+        message -- the commit message, a mandatory parameter
+
         timestamp -- if not None, seconds-since-epoch for a
              postdated/predated commit.
 
@@ -175,45 +215,94 @@ class Commit(object):
 
         allow_pointless -- If true (default), commit even if nothing
             has changed and no merges are recorded.
+
+        strict -- If true, don't allow a commit if the working tree
+            contains unknown files.
+
+        revprops -- Properties for new revision
+        :param local: Perform a local only commit.
         """
         mutter('preparing to commit')
 
-        self.branch = branch
-        self.weave_store = branch.weave_store
+        if deprecated_passed(branch):
+            warn("Commit.commit (branch, ...): The branch parameter is "
+                 "deprecated as of bzr 0.8. Please use working_tree= instead.",
+                 DeprecationWarning, stacklevel=2)
+            self.branch = branch
+            self.work_tree = self.branch.bzrdir.open_workingtree()
+        elif working_tree is None:
+            raise BzrError("One of branch and working_tree must be passed into commit().")
+        else:
+            self.work_tree = working_tree
+            self.branch = self.work_tree.branch
+        if message is None:
+            raise BzrError("The message keyword parameter is required for commit().")
+
+        self.weave_store = self.branch.repository.weave_store
+        self.bound_branch = None
+        self.local = local
+        self.master_branch = None
+        self.master_locked = False
         self.rev_id = rev_id
         self.specific_files = specific_files
         self.allow_pointless = allow_pointless
+        self.revprops = {}
+        if revprops is not None:
+            self.revprops.update(revprops)
 
-        if timestamp is None:
-            self.timestamp = time.time()
-        else:
-            self.timestamp = long(timestamp)
-            
-        if rev_id is None:
-            self.rev_id = _gen_revision_id(self.branch, self.timestamp)
-        else:
-            self.rev_id = rev_id
+        if reporter is None and self.reporter is None:
+            self.reporter = NullCommitReporter()
+        elif reporter is not None:
+            self.reporter = reporter
 
-        if committer is None:
-            self.committer = username(self.branch)
-        else:
-            assert isinstance(committer, basestring), type(committer)
-            self.committer = committer
-
-        if timezone is None:
-            self.timezone = local_time_offset()
-        else:
-            self.timezone = int(timezone)
-
-        assert isinstance(message, basestring), type(message)
-        self.message = message
-        self._escape_commit_message()
-
-        self.branch.lock_write()
+        self.work_tree.lock_write()
         try:
-            self.work_tree = self.branch.working_tree()
+            # setup the bound branch variables as needed.
+            self._check_bound_branch()
+
+            # check for out of date working trees
+            # if we are bound, then self.branch is the master branch and this
+            # test is thus all we need.
+            if self.work_tree.last_revision() != self.master_branch.last_revision():
+                raise errors.OutOfDateTree(self.work_tree)
+    
+            if strict:
+                # raise an exception as soon as we find a single unknown.
+                for unknown in self.work_tree.unknowns():
+                    raise StrictCommitFailed()
+    
+            if timestamp is None:
+                self.timestamp = time.time()
+            else:
+                self.timestamp = long(timestamp)
+                
+            if self.config is None:
+                self.config = bzrlib.config.BranchConfig(self.branch)
+    
+            if rev_id is None:
+                self.rev_id = _gen_revision_id(self.config, self.timestamp)
+            else:
+                self.rev_id = rev_id
+    
+            if committer is None:
+                self.committer = self.config.username()
+            else:
+                assert isinstance(committer, basestring), type(committer)
+                self.committer = committer
+    
+            if timezone is None:
+                self.timezone = local_time_offset()
+            else:
+                self.timezone = int(timezone)
+    
+            if isinstance(message, str):
+                message = message.decode(bzrlib.user_encoding)
+            assert isinstance(message, unicode), type(message)
+            self.message = message
+            self._escape_commit_message()
+
             self.work_inv = self.work_tree.inventory
-            self.basis_tree = self.branch.basis_tree()
+            self.basis_tree = self.work_tree.basis_tree()
             self.basis_inv = self.basis_tree.inventory
 
             self._gather_parents()
@@ -231,40 +320,122 @@ class Commit(object):
                     or self.new_inv != self.basis_inv):
                 raise PointlessCommit()
 
-            if len(list(self.work_tree.iter_conflicts()))>0:
+            if len(self.work_tree.conflicts())>0:
                 raise ConflictsInTree
 
-            self._record_inventory()
+            self.inv_sha1 = self.branch.repository.add_inventory(
+                self.rev_id,
+                self.new_inv,
+                self.present_parents
+                )
             self._make_revision()
-            self.reporter.completed(self.branch.revno()+1, self.rev_id)
-            self.branch.append_revision(self.rev_id)
-            self.branch.set_pending_merges([])
-        finally:
-            self.branch.unlock()
+            # revision data is in the local branch now.
+            
+            # upload revision data to the master.
+            # this will propogate merged revisions too if needed.
+            if self.bound_branch:
+                self.master_branch.repository.fetch(self.branch.repository,
+                                                    revision_id=self.rev_id)
+                # now the master has the revision data
+                # 'commit' to the master first so a timeout here causes the local
+                # branch to be out of date
+                self.master_branch.append_revision(self.rev_id)
 
-    def _record_inventory(self):
-        """Store the inventory for the new revision."""
-        inv_text = serializer_v5.write_inventory_to_string(self.new_inv)
-        self.inv_sha1 = sha_string(inv_text)
-        s = self.branch.control_weaves
-        s.add_text('inventory', self.rev_id,
-                   split_lines(inv_text), self.present_parents,
-                   self.branch.get_transaction())
+            # and now do the commit locally.
+            self.branch.append_revision(self.rev_id)
+
+            self.work_tree.set_pending_merges([])
+            self.work_tree.set_last_revision(self.rev_id)
+            # now the work tree is up to date with the branch
+            
+            self.reporter.completed(self.branch.revno(), self.rev_id)
+            if self.config.post_commit() is not None:
+                hooks = self.config.post_commit().split(' ')
+                # this would be nicer with twisted.python.reflect.namedAny
+                for hook in hooks:
+                    result = eval(hook + '(branch, rev_id)',
+                                  {'branch':self.branch,
+                                   'bzrlib':bzrlib,
+                                   'rev_id':self.rev_id})
+        finally:
+            self._cleanup_bound_branch()
+            self.work_tree.unlock()
+
+    def _check_bound_branch(self):
+        """Check to see if the local branch is bound.
+
+        If it is bound, then most of the commit will actually be
+        done using the remote branch as the target branch.
+        Only at the end will the local branch be updated.
+        """
+        if self.local and not self.branch.get_bound_location():
+            raise errors.LocalRequiresBoundBranch()
+
+        if not self.local:
+            self.master_branch = self.branch.get_master_branch()
+
+        if not self.master_branch:
+            # make this branch the reference branch for out of date checks.
+            self.master_branch = self.branch
+            return
+
+        # If the master branch is bound, we must fail
+        master_bound_location = self.master_branch.get_bound_location()
+        if master_bound_location:
+            raise errors.CommitToDoubleBoundBranch(self.branch,
+                    self.master_branch, master_bound_location)
+
+        # TODO: jam 20051230 We could automatically push local
+        #       commits to the remote branch if they would fit.
+        #       But for now, just require remote to be identical
+        #       to local.
+        
+        # Make sure the local branch is identical to the master
+        master_rh = self.master_branch.revision_history()
+        local_rh = self.branch.revision_history()
+        if local_rh != master_rh:
+            raise errors.BoundBranchOutOfDate(self.branch,
+                    self.master_branch)
+
+        # Now things are ready to change the master branch
+        # so grab the lock
+        self.bound_branch = self.branch
+        self.master_branch.lock_write()
+        self.master_locked = True
+####        
+####        # Check to see if we have any pending merges. If we do
+####        # those need to be pushed into the master branch
+####        pending_merges = self.work_tree.pending_merges()
+####        if pending_merges:
+####            for revision_id in pending_merges:
+####                self.master_branch.repository.fetch(self.bound_branch.repository,
+####                                                    revision_id=revision_id)
+
+    def _cleanup_bound_branch(self):
+        """Executed at the end of a try/finally to cleanup a bound branch.
+
+        If the branch wasn't bound, this is a no-op.
+        If it was, it resents self.branch to the local branch, instead
+        of being the master.
+        """
+        if not self.bound_branch:
+            return
+        if self.master_locked:
+            self.master_branch.unlock()
 
     def _escape_commit_message(self):
         """Replace xml-incompatible control characters."""
+        # FIXME: RBC 20060419 this should be done by the revision
+        # serialiser not by commit. Then we can also add an unescaper
+        # in the deserializer and start roundtripping revision messages
+        # precisely. See repository_implementations/test_repository.py
+        
         # Python strings can include characters that can't be
         # represented in well-formed XML; escape characters that
         # aren't listed in the XML specification
         # (http://www.w3.org/TR/REC-xml/#NT-Char).
-        if isinstance(self.message, unicode):
-            char_pattern = u'[^\x09\x0A\x0D\u0020-\uD7FF\uE000-\uFFFD]'
-        else:
-            # Use a regular 'str' as pattern to avoid having re.subn
-            # return 'unicode' results.
-            char_pattern = '[^x09\x0A\x0D\x20-\xFF]'
         self.message, escape_count = re.subn(
-            char_pattern,
+            u'[^\x09\x0A\x0D\u0020-\uD7FF\uE000-\uFFFD]+',
             lambda match: match.group(0).encode('unicode_escape'),
             self.message)
         if escape_count:
@@ -272,7 +443,7 @@ class Commit(object):
 
     def _gather_parents(self):
         """Record the parents of a merge for merge detection."""
-        pending_merges = self.branch.pending_merges()
+        pending_merges = self.work_tree.pending_merges()
         self.parents = []
         self.parent_invs = []
         self.present_parents = []
@@ -281,14 +452,15 @@ class Commit(object):
             self.parents.append(precursor_id)
         self.parents += pending_merges
         for revision in self.parents:
-            if self.branch.has_revision(revision):
-                self.parent_invs.append(self.branch.get_inventory(revision))
+            if self.branch.repository.has_revision(revision):
+                inventory = self.branch.repository.get_inventory(revision)
+                self.parent_invs.append(inventory)
                 self.present_parents.append(revision)
 
     def _check_parents_present(self):
         for parent_id in self.parents:
             mutter('commit parent revision {%s}', parent_id)
-            if not self.branch.has_revision(parent_id):
+            if not self.branch.repository.has_revision(parent_id):
                 if parent_id == self.branch.last_revision():
                     warning("parent is missing %r", parent_id)
                     raise HistoryMissing(self.branch, 'revision', parent_id)
@@ -297,18 +469,15 @@ class Commit(object):
             
     def _make_revision(self):
         """Record a new revision object for this commit."""
-        self.rev = Revision(timestamp=self.timestamp,
-                            timezone=self.timezone,
-                            committer=self.committer,
-                            message=self.message,
-                            inventory_sha1=self.inv_sha1,
-                            revision_id=self.rev_id)
-        self.rev.parent_ids = self.parents
-        rev_tmp = StringIO()
-        serializer_v5.write_revision(self.rev, rev_tmp)
-        rev_tmp.seek(0)
-        self.branch.revision_store.add(rev_tmp, self.rev_id)
-        mutter('new revision_id is {%s}', self.rev_id)
+        rev = Revision(timestamp=self.timestamp,
+                       timezone=self.timezone,
+                       committer=self.committer,
+                       message=self.message,
+                       inventory_sha1=self.inv_sha1,
+                       revision_id=self.rev_id,
+                       properties=self.revprops)
+        rev.parent_ids = self.parents
+        self.branch.repository.add_revision(self.rev_id, rev, self.new_inv, self.config)
 
     def _remove_deleted(self):
         """Remove deleted files from the working inventories.
@@ -333,7 +502,7 @@ class Commit(object):
             deleted_ids.sort(reverse=True)
             for path, file_id in deleted_ids:
                 del self.work_inv[file_id]
-            self.branch._write_inventory(self.work_inv)
+            self.work_tree._write_inventory(self.work_inv)
 
     def _store_snapshot(self):
         """Pass over inventory and record a snapshot.
@@ -347,16 +516,28 @@ class Commit(object):
         # mark-merge.  
         for path, ie in self.new_inv.iter_entries():
             previous_entries = ie.find_previous_heads(
-                self.parent_invs, 
-                self.weave_store.get_weave_or_empty(ie.file_id,
-                    self.branch.get_transaction()))
+                self.parent_invs,
+                self.weave_store,
+                self.branch.repository.get_transaction())
             if ie.revision is None:
-                change = ie.snapshot(self.rev_id, path, previous_entries,
-                                     self.work_tree, self.weave_store,
-                                     self.branch.get_transaction())
+                # we are creating a new revision for ie in the history store
+                # and inventory.
+                ie.snapshot(self.rev_id, path, previous_entries,
+                    self.work_tree, self.weave_store,
+                    self.branch.repository.get_transaction())
+            # describe the nature of the change that has occured relative to
+            # the basis inventory.
+            if (self.basis_inv.has_id(ie.file_id)):
+                basis_ie = self.basis_inv[ie.file_id]
             else:
-                change = "unchanged"
-            self.reporter.snapshot_change(change, path)
+                basis_ie = None
+            change = ie.describe_change(basis_ie, ie)
+            if change in (InventoryEntry.RENAMED, 
+                InventoryEntry.MODIFIED_AND_RENAMED):
+                old_path = self.basis_inv.id2path(ie.file_id)
+                self.reporter.renamed(change, old_path, path)
+            else:
+                self.reporter.snapshot_change(change, path)
 
     def _populate_new_inv(self):
         """Build revision inventory.
@@ -369,7 +550,7 @@ class Commit(object):
         revision set to their prior value.
         """
         mutter("Selecting files for commit with filter %s", self.specific_files)
-        self.new_inv = Inventory()
+        self.new_inv = Inventory(revision_id=self.rev_id)
         for path, new_ie in self.work_inv.iter_entries():
             file_id = new_ie.file_id
             mutter('check %s {%s}', path, new_ie.file_id)
@@ -409,12 +590,12 @@ class Commit(object):
             self.new_inv.add(self.basis_inv[file_id].copy())
 
     def _report_deletes(self):
-        for file_id in self.basis_inv:
-            if file_id not in self.new_inv:
-                self.reporter.deleted(self.basis_inv.id2path(file_id))
+        for path, ie in self.basis_inv.iter_entries():
+            if ie.file_id not in self.new_inv:
+                self.reporter.deleted(path)
 
-def _gen_revision_id(branch, when):
+def _gen_revision_id(config, when):
     """Return new revision-id."""
-    s = '%s-%s-' % (user_email(branch), compact_date(when))
+    s = '%s-%s-' % (config.user_email(), compact_date(when))
     s += hexlify(rand_bytes(8))
     return s
