@@ -1,4 +1,4 @@
-# -*- coding: UTF-8 -*-
+# Copyright (C) 2004, 2005, 2006 Canonical Ltd.
 
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -14,15 +14,19 @@
 # along with this program; if not, write to the Free Software
 # Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 
-from bzrlib.trace import mutter
-from bzrlib.errors import BzrError
 from bzrlib.delta import compare_trees
+from bzrlib.errors import BzrError
+import bzrlib.errors as errors
+from bzrlib.symbol_versioning import *
+from bzrlib.textfile import check_text_lines
+from bzrlib.trace import mutter
 
 # TODO: Rather than building a changeset object, we should probably
 # invoke callbacks on an object.  That object can either accumulate a
 # list, write them out directly, etc etc.
 
-def internal_diff(old_label, oldlines, new_label, newlines, to_file):
+def internal_diff(old_filename, oldlines, new_filename, newlines, to_file,
+                  allow_binary=False):
     import difflib
     
     # FIXME: difflib is wrong if there is no trailing newline.
@@ -40,9 +44,14 @@ def internal_diff(old_label, oldlines, new_label, newlines, to_file):
     # both sequences are empty.
     if not oldlines and not newlines:
         return
+    
+    if allow_binary is False:
+        check_text_lines(oldlines)
+        check_text_lines(newlines)
 
     ud = difflib.unified_diff(oldlines, newlines,
-                              fromfile=old_label, tofile=new_label)
+                              fromfile=old_filename+'\t', 
+                              tofile=new_filename+'\t')
 
     ud = list(ud)
     # work-around for difflib being too smart for its own good
@@ -62,7 +71,7 @@ def internal_diff(old_label, oldlines, new_label, newlines, to_file):
     print >>to_file
 
 
-def external_diff(old_label, oldlines, new_label, newlines, to_file,
+def external_diff(old_filename, oldlines, new_filename, newlines, to_file,
                   diff_opts):
     """Display a diff by calling out to the external diff program."""
     import sys
@@ -97,9 +106,9 @@ def external_diff(old_label, oldlines, new_label, newlines, to_file,
         if not diff_opts:
             diff_opts = []
         diffcmd = ['diff',
-                   '--label', old_label,
+                   '--label', old_filename+'\t',
                    oldtmpf.name,
-                   '--label', new_label,
+                   '--label', new_filename+'\t',
                    newtmpf.name]
 
         # diff only allows one style to be specified; they don't override.
@@ -140,10 +149,14 @@ def external_diff(old_label, oldlines, new_label, newlines, to_file,
         oldtmpf.close()                 # and delete
         newtmpf.close()
 
+
+@deprecated_function(zero_eight)
 def show_diff(b, from_spec, specific_files, external_diff_options=None,
-              revision2=None, output=None):
+              revision2=None, output=None, b2=None):
     """Shortcut for showing the diff to the working tree.
 
+    Please use show_diff_trees instead.
+
     b
         Branch.
 
@@ -158,22 +171,72 @@ def show_diff(b, from_spec, specific_files, external_diff_options=None,
         output = sys.stdout
 
     if from_spec is None:
-        old_tree = b.basis_tree()
+        old_tree = b.bzrdir.open_workingtree()
+        if b2 is None:
+            old_tree = old_tree = old_tree.basis_tree()
     else:
-        old_tree = b.revision_tree(from_spec.in_history(b).rev_id)
+        old_tree = b.repository.revision_tree(from_spec.in_history(b).rev_id)
 
     if revision2 is None:
-        new_tree = b.working_tree()
+        if b2 is None:
+            new_tree = b.bzrdir.open_workingtree()
+        else:
+            new_tree = b2.bzrdir.open_workingtree()
     else:
-        new_tree = b.revision_tree(revision2.in_history(b).rev_id)
+        new_tree = b.repository.revision_tree(revision2.in_history(b).rev_id)
 
-    show_diff_trees(old_tree, new_tree, output, specific_files,
-                    external_diff_options)
+    return show_diff_trees(old_tree, new_tree, output, specific_files,
+                           external_diff_options)
 
 
+def diff_cmd_helper(tree, specific_files, external_diff_options, 
+                    old_revision_spec=None, new_revision_spec=None,
+                    old_label='a/', new_label='b/'):
+    """Helper for cmd_diff.
+
+   tree 
+        A WorkingTree
+
+    specific_files
+        The specific files to compare, or None
+
+    external_diff_options
+        If non-None, run an external diff, and pass it these options
+
+    old_revision_spec
+        If None, use basis tree as old revision, otherwise use the tree for
+        the specified revision. 
+
+    new_revision_spec
+        If None, use working tree as new revision, otherwise use the tree for
+        the specified revision.
+    
+    The more general form is show_diff_trees(), where the caller
+    supplies any two trees.
+    """
+    import sys
+    output = sys.stdout
+    def spec_tree(spec):
+        revision_id = spec.in_store(tree.branch).rev_id
+        return tree.branch.repository.revision_tree(revision_id)
+    if old_revision_spec is None:
+        old_tree = tree.basis_tree()
+    else:
+        old_tree = spec_tree(old_revision_spec)
+
+    if new_revision_spec is None:
+        new_tree = tree
+    else:
+        new_tree = spec_tree(new_revision_spec)
+
+    return show_diff_trees(old_tree, new_tree, sys.stdout, specific_files,
+                           external_diff_options,
+                           old_label=old_label, new_label=new_label)
+
 
 def show_diff_trees(old_tree, new_tree, to_file, specific_files=None,
-                    external_diff_options=None):
+                    external_diff_options=None,
+                    old_label='a/', new_label='b/'):
     """Show in text form the changes from one tree to another.
 
     to_files
@@ -182,10 +245,22 @@ def show_diff_trees(old_tree, new_tree, to_file, specific_files=None,
     external_diff_options
         If set, use an external GNU diff and pass these options.
     """
+    old_tree.lock_read()
+    try:
+        new_tree.lock_read()
+        try:
+            return _show_diff_trees(old_tree, new_tree, to_file,
+                                    specific_files, external_diff_options,
+                                    old_label=old_label, new_label=new_label)
+        finally:
+            new_tree.unlock()
+    finally:
+        old_tree.unlock()
 
-    # TODO: Options to control putting on a prefix or suffix, perhaps as a format string
-    old_label = ''
-    new_label = ''
+
+def _show_diff_trees(old_tree, new_tree, to_file,
+                     specific_files, external_diff_options, 
+                     old_label='a/', new_label='b/' ):
 
     DEVNULL = '/dev/null'
     # Windows users, don't panic about this filename -- it is a
@@ -195,6 +270,8 @@ def show_diff_trees(old_tree, new_tree, to_file, specific_files=None,
     # TODO: Generation of pseudo-diffs for added/deleted files could
     # be usefully made into a much faster special case.
 
+    _raise_if_doubly_unversioned(specific_files, old_tree, new_tree)
+
     if external_diff_options:
         assert isinstance(external_diff_options, basestring)
         opts = external_diff_options.split()
@@ -203,36 +280,72 @@ def show_diff_trees(old_tree, new_tree, to_file, specific_files=None,
     else:
         diff_file = internal_diff
     
-
     delta = compare_trees(old_tree, new_tree, want_unchanged=False,
                           specific_files=specific_files)
 
+    has_changes = 0
     for path, file_id, kind in delta.removed:
+        has_changes = 1
         print >>to_file, '=== removed %s %r' % (kind, path)
         old_tree.inventory[file_id].diff(diff_file, old_label + path, old_tree,
                                          DEVNULL, None, None, to_file)
     for path, file_id, kind in delta.added:
+        has_changes = 1
         print >>to_file, '=== added %s %r' % (kind, path)
         new_tree.inventory[file_id].diff(diff_file, new_label + path, new_tree,
                                          DEVNULL, None, None, to_file, 
                                          reverse=True)
     for (old_path, new_path, file_id, kind,
          text_modified, meta_modified) in delta.renamed:
+        has_changes = 1
         prop_str = get_prop_change(meta_modified)
         print >>to_file, '=== renamed %s %r => %r%s' % (
-                          kind, old_path, new_path, prop_str)
+                    kind, old_path, new_path, prop_str)
         _maybe_diff_file_or_symlink(old_label, old_path, old_tree, file_id,
                                     new_label, new_path, new_tree,
                                     text_modified, kind, to_file, diff_file)
     for path, file_id, kind, text_modified, meta_modified in delta.modified:
+        has_changes = 1
         prop_str = get_prop_change(meta_modified)
         print >>to_file, '=== modified %s %r%s' % (kind, path, prop_str)
         if text_modified:
             _maybe_diff_file_or_symlink(old_label, path, old_tree, file_id,
                                         new_label, path, new_tree,
                                         True, kind, to_file, diff_file)
+
+    return has_changes
+
+
+def _raise_if_doubly_unversioned(specific_files, old_tree, new_tree):
+    """Complain if paths are not versioned in either tree."""
+    if not specific_files:
+        return
+    old_unversioned = old_tree.filter_unversioned_files(specific_files)
+    new_unversioned = new_tree.filter_unversioned_files(specific_files)
+    unversioned = old_unversioned.intersection(new_unversioned)
+    if unversioned:
+        raise errors.PathsNotVersionedError(sorted(unversioned))
     
 
+def _raise_if_nonexistent(paths, old_tree, new_tree):
+    """Complain if paths are not in either inventory or tree.
+
+    It's OK with the files exist in either tree's inventory, or 
+    if they exist in the tree but are not versioned.
+    
+    This can be used by operations such as bzr status that can accept
+    unknown or ignored files.
+    """
+    mutter("check paths: %r", paths)
+    if not paths:
+        return
+    s = old_tree.filter_unversioned_files(paths)
+    s = new_tree.filter_unversioned_files(s)
+    s = [path for path in s if not new_tree.has_filename(path)]
+    if s:
+        raise errors.PathsDoNotExist(sorted(s))
+
+
 def get_prop_change(meta_modified):
     if meta_modified:
         return " (properties changed)"
