@@ -1,19 +1,23 @@
-# (C) 2005 Canonical Ltd
-
+# Copyright (C) 2005, 2006 Canonical Ltd
+#
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
 # the Free Software Foundation; either version 2 of the License, or
 # (at your option) any later version.
-
+#
 # This program is distributed in the hope that it will be useful,
 # but WITHOUT ANY WARRANTY; without even the implied warranty of
 # MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 # GNU General Public License for more details.
-
+#
 # You should have received a copy of the GNU General Public License
 # along with this program; if not, write to the Free Software
 # Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 
+# FIXME: This refactoring of the workingtree code doesn't seem to keep 
+# the WorkingTree's copy of the inventory in sync with the branch.  The
+# branch modifies its working inventory when it does a commit to make
+# missing files permanently removed.
 
 # TODO: Maybe also keep the full path of the entry, and the children?
 # But those depend on its position within a particular inventory, and
@@ -31,12 +35,11 @@ import tarfile
 import types
 
 import bzrlib
-from bzrlib.errors import BzrError, BzrCheckError
-
 from bzrlib.osutils import (pumpfile, quotefn, splitpath, joinpath,
-                            appendpath, sha_strings)
+                            pathjoin, sha_strings)
+from bzrlib.errors import (NotVersionedError, InvalidEntryName,
+                           BzrError, BzrCheckError, BinaryFile)
 from bzrlib.trace import mutter
-from bzrlib.errors import NotVersionedError
 
 
 class InventoryEntry(object):
@@ -76,11 +79,12 @@ class InventoryEntry(object):
     InventoryDirectory('123', 'src', parent_id='TREE_ROOT')
     >>> i.add(InventoryFile('2323', 'hello.c', parent_id='123'))
     InventoryFile('2323', 'hello.c', parent_id='123')
-    >>> for j in i.iter_entries():
-    ...   print j
+    >>> shouldbe = {0: 'src', 1: pathjoin('src','hello.c')}
+    >>> for ix, j in enumerate(i.iter_entries()):
+    ...   print (j[0] == shouldbe[ix], j[1])
     ... 
-    ('src', InventoryDirectory('123', 'src', parent_id='TREE_ROOT'))
-    ('src/hello.c', InventoryFile('2323', 'hello.c', parent_id='123'))
+    (True, InventoryDirectory('123', 'src', parent_id='TREE_ROOT'))
+    (True, InventoryFile('2323', 'hello.c', parent_id='123'))
     >>> i.add(InventoryFile('2323', 'bye.c', '123'))
     Traceback (most recent call last):
     ...
@@ -98,7 +102,7 @@ class InventoryEntry(object):
     >>> i['2326']
     InventoryFile('2326', 'wibble.c', parent_id='2325')
     >>> for path, entry in i.iter_entries():
-    ...     print path.replace('\\\\', '/')     # for win32 os.sep
+    ...     print path
     ...     assert i.path2id(path)
     ... 
     src
@@ -106,17 +110,27 @@ class InventoryEntry(object):
     src/hello.c
     src/wibble
     src/wibble/wibble.c
-    >>> i.id2path('2326').replace('\\\\', '/')
+    >>> i.id2path('2326')
     'src/wibble/wibble.c'
     """
+
+    # Constants returned by describe_change()
+    #
+    # TODO: These should probably move to some kind of FileChangeDescription 
+    # class; that's like what's inside a TreeDelta but we want to be able to 
+    # generate them just for one file at a time.
+    RENAMED = 'renamed'
+    MODIFIED_AND_RENAMED = 'modified and renamed'
     
     __slots__ = ['text_sha1', 'text_size', 'file_id', 'name', 'kind',
                  'text_id', 'parent_id', 'children', 'executable', 
                  'revision']
 
     def _add_text_to_weave(self, new_lines, parents, weave_store, transaction):
-        weave_store.add_text(self.file_id, self.revision, new_lines, parents,
-                             transaction)
+        versionedfile = weave_store.get_weave_or_empty(self.file_id,
+                                                       transaction)
+        versionedfile.add_lines(self.revision, parents, new_lines)
+        versionedfile.clear_cache()
 
     def detect_changes(self, old_entry):
         """Return a (text_modified, meta_modified) from this to old_entry.
@@ -147,7 +161,10 @@ class InventoryEntry(object):
              output_to, reverse=False):
         """Perform a diff between two entries of the same kind."""
 
-    def find_previous_heads(self, previous_inventories, entry_weave):
+    def find_previous_heads(self, previous_inventories,
+                            versioned_file_store,
+                            transaction,
+                            entry_vf=None):
         """Return the revisions and entries that directly preceed this.
 
         Returned as a map from revision to inventory entry.
@@ -155,50 +172,84 @@ class InventoryEntry(object):
         This is a map containing the file revisions in all parents
         for which the file exists, and its revision is not a parent of
         any other. If the file is new, the set will be empty.
+
+        :param versioned_file_store: A store where ancestry data on this
+                                     file id can be queried.
+        :param transaction: The transaction that queries to the versioned 
+                            file store should be completed under.
+        :param entry_vf: The entry versioned file, if its already available.
         """
         def get_ancestors(weave, entry):
-            return set(map(weave.idx_to_name,
-                           weave.inclusions([weave.lookup(entry.revision)])))
+            return set(weave.get_ancestry(entry.revision))
+        # revision:ie mapping for each ie found in previous_inventories.
+        candidates = {}
+        # revision:ie mapping with one revision for each head.
         heads = {}
+        # revision: ancestor list for each head
         head_ancestors = {}
+        # identify candidate head revision ids.
         for inv in previous_inventories:
             if self.file_id in inv:
                 ie = inv[self.file_id]
                 assert ie.file_id == self.file_id
-                if ie.revision in heads:
-                    # fixup logic, there was a bug in revision updates.
-                    # with x bit support.
+                if ie.revision in candidates:
+                    # same revision value in two different inventories:
+                    # correct possible inconsistencies:
+                    #     * there was a bug in revision updates with 'x' bit 
+                    #       support.
                     try:
-                        if heads[ie.revision].executable != ie.executable:
-                            heads[ie.revision].executable = False
+                        if candidates[ie.revision].executable != ie.executable:
+                            candidates[ie.revision].executable = False
                             ie.executable = False
                     except AttributeError:
                         pass
-                    assert heads[ie.revision] == ie
+                    # must now be the same.
+                    assert candidates[ie.revision] == ie
                 else:
-                    # may want to add it.
-                    # may already be covered:
-                    already_present = 0 != len(
-                        [head for head in heads 
-                         if ie.revision in head_ancestors[head]])
-                    if already_present:
-                        # an ancestor of a known head.
-                        continue
-                    # definately a head:
-                    ancestors = get_ancestors(entry_weave, ie)
-                    # may knock something else out:
-                    check_heads = list(heads.keys())
-                    for head in check_heads:
-                        if head in ancestors:
-                            # this head is not really a head
-                            heads.pop(head)
-                    head_ancestors[ie.revision] = ancestors
-                    heads[ie.revision] = ie
+                    # add this revision as a candidate.
+                    candidates[ie.revision] = ie
+
+        # common case optimisation
+        if len(candidates) == 1:
+            # if there is only one candidate revision found
+            # then we can opening the versioned file to access ancestry:
+            # there cannot be any ancestors to eliminate when there is 
+            # only one revision available.
+            heads[ie.revision] = ie
+            return heads
+
+        # eliminate ancestors amongst the available candidates:
+        # heads are those that are not an ancestor of any other candidate
+        # - this provides convergence at a per-file level.
+        for ie in candidates.values():
+            # may be an ancestor of a known head:
+            already_present = 0 != len(
+                [head for head in heads 
+                 if ie.revision in head_ancestors[head]])
+            if already_present:
+                # an ancestor of an analyzed candidate.
+                continue
+            # not an ancestor of a known head:
+            # load the versioned file for this file id if needed
+            if entry_vf is None:
+                entry_vf = versioned_file_store.get_weave_or_empty(
+                    self.file_id, transaction)
+            ancestors = get_ancestors(entry_vf, ie)
+            # may knock something else out:
+            check_heads = list(heads.keys())
+            for head in check_heads:
+                if head in ancestors:
+                    # this previously discovered 'head' is not
+                    # really a head - its an ancestor of the newly 
+                    # found head,
+                    heads.pop(head)
+            head_ancestors[ie.revision] = ancestors
+            heads[ie.revision] = ie
         return heads
 
     def get_tar_item(self, root, dp, now, tree):
         """Get a tarfile item and a file stream for its content."""
-        item = tarfile.TarInfo(os.path.join(root, dp))
+        item = tarfile.TarInfo(pathjoin(root, dp))
         # TODO: would be cool to actually set it to the timestamp of the
         # revision it was last changed
         item.mtime = now
@@ -229,12 +280,11 @@ class InventoryEntry(object):
         '123'
         >>> e = InventoryFile('123', 'src/hello.c', ROOT_ID)
         Traceback (most recent call last):
-        BzrCheckError: InventoryEntry name 'src/hello.c' is invalid
+        InvalidEntryName: Invalid entry name: src/hello.c
         """
         assert isinstance(name, basestring), name
         if '/' in name or '\\' in name:
-            raise BzrCheckError('InventoryEntry name %r is invalid' % name)
-        
+            raise InvalidEntryName(name=name)
         self.executable = False
         self.revision = None
         self.text_sha1 = None
@@ -264,9 +314,10 @@ class InventoryEntry(object):
         
         This is a template method - implement _put_on_disk in subclasses.
         """
-        fullpath = appendpath(dest, dp)
+        fullpath = pathjoin(dest, dp)
         self._put_on_disk(fullpath, tree)
-        mutter("  export {%s} kind %s to %s" % (self.file_id, self.kind, fullpath))
+        mutter("  export {%s} kind %s to %s", self.file_id,
+                self.kind, fullpath)
 
     def _put_on_disk(self, fullpath, tree):
         """Put this entry onto disk at fullpath, from tree tree."""
@@ -286,6 +337,14 @@ class InventoryEntry(object):
 
         This is a template method, override _check for kind specific
         tests.
+
+        :param checker: Check object providing context for the checks; 
+             can be used to find out what parts of the repository have already
+             been checked.
+        :param rev_id: Revision id from which this InventoryEntry was loaded.
+             Not necessarily the last-changed revision for this file.
+        :param inv: Inventory from which the entry was loaded.
+        :param tree: RevisionTree for this entry.
         """
         if self.parent_id != None:
             if not inv.has_id(self.parent_id):
@@ -298,18 +357,49 @@ class InventoryEntry(object):
         raise BzrCheckError('unknown entry kind %r in revision {%s}' % 
                             (self.kind, rev_id))
 
-
     def copy(self):
         """Clone this inventory entry."""
         raise NotImplementedError
 
-    def _get_snapshot_change(self, previous_entries):
-        if len(previous_entries) > 1:
-            return 'merged'
-        elif len(previous_entries) == 0:
+    @staticmethod
+    def describe_change(old_entry, new_entry):
+        """Describe the change between old_entry and this.
+        
+        This smells of being an InterInventoryEntry situation, but as its
+        the first one, we're making it a static method for now.
+
+        An entry with a different parent, or different name is considered 
+        to be renamed. Reparenting is an internal detail.
+        Note that renaming the parent does not trigger a rename for the
+        child entry itself.
+        """
+        # TODO: Perhaps return an object rather than just a string
+        if old_entry is new_entry:
+            # also the case of both being None
+            return 'unchanged'
+        elif old_entry is None:
             return 'added'
+        elif new_entry is None:
+            return 'removed'
+        text_modified, meta_modified = new_entry.detect_changes(old_entry)
+        if text_modified or meta_modified:
+            modified = True
         else:
-            return 'modified/renamed/reparented'
+            modified = False
+        # TODO 20060511 (mbp, rbc) factor out 'detect_rename' here.
+        if old_entry.parent_id != new_entry.parent_id:
+            renamed = True
+        elif old_entry.name != new_entry.name:
+            renamed = True
+        else:
+            renamed = False
+        if renamed and not modified:
+            return InventoryEntry.RENAMED
+        if modified and not renamed:
+            return 'modified'
+        if modified and renamed:
+            return InventoryEntry.MODIFIED_AND_RENAMED
+        return 'unchanged'
 
     def __repr__(self):
         return ("%s(%r, %r, parent_id=%r)"
@@ -334,18 +424,24 @@ class InventoryEntry(object):
                 mutter("found unchanged entry")
                 self.revision = parent_ie.revision
                 return "unchanged"
-        return self.snapshot_revision(revision, previous_entries, 
-                                      work_tree, weave_store, transaction)
+        return self._snapshot_into_revision(revision, previous_entries, 
+                                            work_tree, weave_store, transaction)
 
-    def snapshot_revision(self, revision, previous_entries, work_tree,
-                          weave_store, transaction):
-        """Record this revision unconditionally."""
-        mutter('new revision for {%s}', self.file_id)
+    def _snapshot_into_revision(self, revision, previous_entries, work_tree,
+                                weave_store, transaction):
+        """Record this revision unconditionally into a store.
+
+        The entry's last-changed revision property (`revision`) is updated to 
+        that of the new revision.
+        
+        :param revision: id of the new revision that is being recorded.
+
+        :returns: String description of the commit (e.g. "merged", "modified"), etc.
+        """
+        mutter('new revision {%s} for {%s}', revision, self.file_id)
         self.revision = revision
-        change = self._get_snapshot_change(previous_entries)
         self._snapshot_text(previous_entries, work_tree, weave_store,
                             transaction)
-        return change
 
     def _snapshot_text(self, file_parents, work_tree, weave_store, transaction): 
         """Record the 'text' of this entry, whatever form that takes.
@@ -354,7 +450,7 @@ class InventoryEntry(object):
         """
         mutter('storing file {%s} in revision {%s}',
                self.file_id, self.revision)
-        self._add_text_to_weave([], file_parents, weave_store, transaction)
+        self._add_text_to_weave([], file_parents.keys(), weave_store, transaction)
 
     def __eq__(self, other):
         if not isinstance(other, InventoryEntry):
@@ -403,6 +499,9 @@ class InventoryEntry(object):
         # first requested, or preload them if they're already known
         pass            # nothing to do by default
 
+    def _forget_tree_state(self):
+        pass
+
 
 class RootEntry(InventoryEntry):
 
@@ -414,7 +513,7 @@ class RootEntry(InventoryEntry):
         self.children = {}
         self.kind = 'root_directory'
         self.parent_id = None
-        self.name = ''
+        self.name = u''
 
     def __eq__(self, other):
         if not isinstance(other, RootEntry):
@@ -466,25 +565,37 @@ class InventoryDirectory(InventoryEntry):
 class InventoryFile(InventoryEntry):
     """A file in an inventory."""
 
-    def _check(self, checker, rev_id, tree):
+    def _check(self, checker, tree_revision_id, tree):
         """See InventoryEntry._check"""
-        revision = self.revision
-        t = (self.file_id, revision)
+        t = (self.file_id, self.revision)
         if t in checker.checked_texts:
-            prev_sha = checker.checked_texts[t] 
+            prev_sha = checker.checked_texts[t]
             if prev_sha != self.text_sha1:
                 raise BzrCheckError('mismatched sha1 on {%s} in {%s}' %
-                                    (self.file_id, rev_id))
+                                    (self.file_id, tree_revision_id))
             else:
                 checker.repeated_text_cnt += 1
                 return
-        mutter('check version {%s} of {%s}', rev_id, self.file_id)
-        file_lines = tree.get_file_lines(self.file_id)
-        checker.checked_text_cnt += 1 
-        if self.text_size != sum(map(len, file_lines)):
-            raise BzrCheckError('text {%s} wrong size' % self.text_id)
-        if self.text_sha1 != sha_strings(file_lines):
-            raise BzrCheckError('text {%s} wrong sha1' % self.text_id)
+
+        if self.file_id not in checker.checked_weaves:
+            mutter('check weave {%s}', self.file_id)
+            w = tree.get_weave(self.file_id)
+            # Not passing a progress bar, because it creates a new
+            # progress, which overwrites the current progress,
+            # and doesn't look nice
+            w.check()
+            checker.checked_weaves[self.file_id] = True
+        else:
+            w = tree.get_weave(self.file_id)
+
+        mutter('check version {%s} of {%s}', tree_revision_id, self.file_id)
+        checker.checked_text_cnt += 1
+        # We can't check the length, because Weave doesn't store that
+        # information, and the whole point of looking at the weave's
+        # sha1sum is that we don't have to extract the text.
+        if self.text_sha1 != w.get_sha1(self.revision):
+            raise BzrCheckError('text {%s} version {%s} wrong sha1' 
+                                % (self.file_id, self.revision))
         checker.checked_texts[t] = self.text_sha1
 
     def copy(self):
@@ -507,17 +618,24 @@ class InventoryFile(InventoryEntry):
     def _diff(self, text_diff, from_label, tree, to_label, to_entry, to_tree,
              output_to, reverse=False):
         """See InventoryEntry._diff."""
-        from_text = tree.get_file(self.file_id).readlines()
-        if to_entry:
-            to_text = to_tree.get_file(to_entry.file_id).readlines()
-        else:
-            to_text = []
-        if not reverse:
-            text_diff(from_label, from_text,
-                      to_label, to_text, output_to)
-        else:
-            text_diff(to_label, to_text,
-                      from_label, from_text, output_to)
+        try:
+            from_text = tree.get_file(self.file_id).readlines()
+            if to_entry:
+                to_text = to_tree.get_file(to_entry.file_id).readlines()
+            else:
+                to_text = []
+            if not reverse:
+                text_diff(from_label, from_text,
+                          to_label, to_text, output_to)
+            else:
+                text_diff(to_label, to_text,
+                          from_label, from_text, output_to)
+        except BinaryFile:
+            if reverse:
+                label_pair = (to_label, from_label)
+            else:
+                label_pair = (from_label, to_label)
+            print >> output_to, "Binary files %s and %s differ" % label_pair
 
     def has_text(self):
         """See InventoryEntry.has_text."""
@@ -553,22 +671,25 @@ class InventoryFile(InventoryEntry):
         self.text_sha1 = work_tree.get_file_sha1(self.file_id)
         self.executable = work_tree.is_executable(self.file_id)
 
-    def _snapshot_text(self, file_parents, work_tree, weave_store, transaction):
+    def _forget_tree_state(self):
+        self.text_sha1 = None
+        self.executable = None
+
+    def _snapshot_text(self, file_parents, work_tree, versionedfile_store, transaction):
         """See InventoryEntry._snapshot_text."""
-        mutter('storing file {%s} in revision {%s}',
-               self.file_id, self.revision)
+        mutter('storing text of file {%s} in revision {%s} into %r',
+               self.file_id, self.revision, versionedfile_store)
         # special case to avoid diffing on renames or 
         # reparenting
         if (len(file_parents) == 1
             and self.text_sha1 == file_parents.values()[0].text_sha1
             and self.text_size == file_parents.values()[0].text_size):
             previous_ie = file_parents.values()[0]
-            weave_store.add_identical_text(
-                self.file_id, previous_ie.revision, 
-                self.revision, file_parents, transaction)
+            versionedfile = versionedfile_store.get_weave(self.file_id, transaction)
+            versionedfile.clone_text(self.revision, previous_ie.revision, file_parents.keys())
         else:
             new_lines = work_tree.get_file(self.file_id).readlines()
-            self._add_text_to_weave(new_lines, file_parents, weave_store,
+            self._add_text_to_weave(new_lines, file_parents.keys(), versionedfile_store,
                                     transaction)
             self.text_sha1 = sha_strings(new_lines)
             self.text_size = sum(map(len, new_lines))
@@ -644,7 +765,7 @@ class InventoryLink(InventoryEntry):
 
     def _put_in_tar(self, item, tree):
         """See InventoryEntry._put_in_tar."""
-        iterm.type = tarfile.SYMTYPE
+        item.type = tarfile.SYMTYPE
         fileobj = None
         item.size = 0
         item.mode = 0755
@@ -662,6 +783,9 @@ class InventoryLink(InventoryEntry):
         """See InventoryEntry._read_tree_state."""
         self.symlink_target = work_tree.get_symlink_target(self.file_id)
 
+    def _forget_tree_state(self):
+        self.symlink_target = None
+
     def _unchanged(self, previous_ie):
         """See InventoryEntry._unchanged."""
         compatible = super(InventoryLink, self)._unchanged(previous_ie)
@@ -708,7 +832,7 @@ class Inventory(object):
     >>> inv.add(InventoryFile('123-123', 'hello.c', ROOT_ID))
     InventoryFile('123-123', 'hello.c', parent_id='TREE_ROOT-12345678-12345678')
     """
-    def __init__(self, root_id=ROOT_ID):
+    def __init__(self, root_id=ROOT_ID, revision_id=None):
         """Create or read an inventory.
 
         If a working directory is specified, the inventory is read
@@ -718,15 +842,17 @@ class Inventory(object):
         The inventory is created with a default root directory, with
         an id of None.
         """
-        # We are letting Branch.initialize() create a unique inventory
+        # We are letting Branch.create() create a unique inventory
         # root id. Rather than generating a random one here.
         #if root_id is None:
         #    root_id = bzrlib.branch.gen_file_id('TREE_ROOT')
         self.root = RootEntry(root_id)
+        self.revision_id = revision_id
         self._byid = {self.root.file_id: self.root}
 
 
     def copy(self):
+        # TODO: jam 20051218 Should copy also copy the revision_id?
         other = Inventory(self.root.file_id)
         # copy recursively so we know directories will be added before
         # their children.  There are more efficient ways than this...
@@ -760,7 +886,7 @@ class Inventory(object):
             yield name, ie
             if ie.kind == 'directory':
                 for cn, cie in self.iter_entries(from_dir=ie.file_id):
-                    yield os.path.join(name, cn), cie
+                    yield pathjoin(name, cn), cie
 
 
     def entries(self):
@@ -773,12 +899,12 @@ class Inventory(object):
             kids = dir_ie.children.items()
             kids.sort()
             for name, ie in kids:
-                child_path = os.path.join(dir_path, name)
+                child_path = pathjoin(dir_path, name)
                 accum.append((child_path, ie))
                 if ie.kind == 'directory':
                     descend(ie, child_path)
 
-        descend(self.root, '')
+        descend(self.root, u'')
         return accum
 
 
@@ -793,9 +919,9 @@ class Inventory(object):
             kids.sort()
 
             for name, child_ie in kids:
-                child_path = os.path.join(parent_path, name)
+                child_path = pathjoin(parent_path, name)
                 descend(child_ie, child_path)
-        descend(self.root, '')
+        descend(self.root, u'')
         return accum
         
 
@@ -860,7 +986,7 @@ class Inventory(object):
 
         if parent.children.has_key(entry.name):
             raise BzrError("%s is already versioned" %
-                    appendpath(self.id2path(parent.file_id), entry.name))
+                    pathjoin(self.id2path(parent.file_id), entry.name))
 
         self._byid[entry.file_id] = entry
         parent.children[entry.name] = entry
@@ -873,20 +999,22 @@ class Inventory(object):
         The immediate parent must already be versioned.
 
         Returns the new entry object."""
-        from bzrlib.branch import gen_file_id
+        from bzrlib.workingtree import gen_file_id
         
         parts = bzrlib.osutils.splitpath(relpath)
-        if len(parts) == 0:
-            raise BzrError("cannot re-add root of inventory")
 
         if file_id == None:
             file_id = gen_file_id(relpath)
 
-        parent_path = parts[:-1]
-        parent_id = self.path2id(parent_path)
-        if parent_id == None:
-            raise NotVersionedError(parent_path)
-
+        if len(parts) == 0:
+            self.root = RootEntry(file_id)
+            self._byid = {self.root.file_id: self.root}
+            return
+        else:
+            parent_path = parts[:-1]
+            parent_id = self.path2id(parent_path)
+            if parent_id == None:
+                raise NotVersionedError(path=parent_path)
         if kind == 'directory':
             ie = InventoryDirectory(file_id, parts[-1], parent_id)
         elif kind == 'file':
@@ -912,17 +1040,12 @@ class Inventory(object):
         """
         ie = self[file_id]
 
-        assert self[ie.parent_id].children[ie.name] == ie
+        assert ie.parent_id is None or \
+            self[ie.parent_id].children[ie.name] == ie
         
-        # TODO: Test deleting all children; maybe hoist to a separate
-        # deltree method?
-        if ie.kind == 'directory':
-            for cie in ie.children.values():
-                del self[cie.file_id]
-            del ie.children
-
         del self._byid[file_id]
-        del self[ie.parent_id].children[ie.name]
+        if ie.parent_id is not None:
+            del self[ie.parent_id].children[ie.name]
 
 
     def __eq__(self, other):
@@ -958,6 +1081,15 @@ class Inventory(object):
     def __hash__(self):
         raise ValueError('not hashable')
 
+    def _iter_file_id_parents(self, file_id):
+        """Yield the parents of file_id up to the root."""
+        while file_id != None:
+            try:
+                ie = self._byid[file_id]
+            except KeyError:
+                raise BzrError("file_id {%s} not found in inventory" % file_id)
+            yield ie
+            file_id = ie.parent_id
 
     def get_idpath(self, file_id):
         """Return a list of file_ids for the path to an entry.
@@ -968,31 +1100,24 @@ class Inventory(object):
         root directory as depth 1.
         """
         p = []
-        while file_id != None:
-            try:
-                ie = self._byid[file_id]
-            except KeyError:
-                raise BzrError("file_id {%s} not found in inventory" % file_id)
-            p.insert(0, ie.file_id)
-            file_id = ie.parent_id
+        for parent in self._iter_file_id_parents(file_id):
+            p.insert(0, parent.file_id)
         return p
 
-
     def id2path(self, file_id):
-        """Return as a list the path to file_id.
+        """Return as a string the path to file_id.
         
         >>> i = Inventory()
         >>> e = i.add(InventoryDirectory('src-id', 'src', ROOT_ID))
         >>> e = i.add(InventoryFile('foo-id', 'foo.c', parent_id='src-id'))
-        >>> print i.id2path('foo-id').replace(os.sep, '/')
+        >>> print i.id2path('foo-id')
         src/foo.c
         """
         # get all names, skipping root
-        p = [self._byid[fid].name for fid in self.get_idpath(file_id)[1:]]
-        return os.sep.join(p)
+        return '/'.join(reversed(
+            [parent.name for parent in 
+             self._iter_file_id_parents(file_id)][:-1]))
             
-
-
     def path2id(self, name):
         """Walk down through directories to return entry of last component.
 
