@@ -1,4 +1,4 @@
-# Copyright (C) 2005 Canonical Ltd
+# Copyright (C) 2005, 2006 Canonical Ltd
 
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -15,35 +15,34 @@
 # Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 
 
-"""Locking wrappers.
+"""Locking using OS file locks or file existence.
 
-This only does local locking using OS locks for now.
+Note: This method of locking is generally deprecated in favour of LockDir, but
+is used to lock local WorkingTrees, and by some old formats.  It's accessed
+through Transport.lock_read(), etc.
 
 This module causes two methods, lock() and unlock() to be defined in
 any way that works on the current platform.
 
 It is not specified whether these locks are reentrant (i.e. can be
 taken repeatedly by a single process) or whether they exclude
-different threads in a single process.  
-
-Eventually we may need to use some kind of lock representation that
-will work on a dumb filesystem without actual locking primitives.
+different threads in a single process.  That reentrancy is provided by 
+LockableFiles.
 
 This defines two classes: ReadLock and WriteLock, which can be
 implemented in different ways on different platforms.  Both have an
 unlock() method.
 """
 
-
-import sys
+import errno
 import os
+import sys
 
 from bzrlib.trace import mutter, note, warning
 from bzrlib.errors import LockError
 
 class _base_Lock(object):
     def _open(self, filename, filemode):
-        import errno
         try:
             self.f = open(filename, filemode)
             return self.f
@@ -52,19 +51,17 @@ class _base_Lock(object):
                 raise
 
             # maybe this is an old branch (before may 2005)
-            mutter("trying to create missing branch lock %r" % filename)
+            mutter("trying to create missing branch lock %r", filename)
             
-            self.f = open(filename, 'wb')
+            self.f = open(filename, 'wb+')
             return self.f
 
-
     def __del__(self):
         if self.f:
             from warnings import warn
             warn("lock on %r not released" % self.f)
             self.unlock()
             
-
     def unlock(self):
         raise NotImplementedError()
 
@@ -88,33 +85,41 @@ try:
             self.f.close()
             del self.f 
 
-
     class _fcntl_WriteLock(_fcntl_FileLock):
         def __init__(self, filename):
+            # standard IO errors get exposed directly.
+            self._open(filename, 'wb')
             try:
-                fcntl.lockf(self._open(filename, 'wb'), fcntl.LOCK_EX)
-            except Exception, e:
+                fcntl.lockf(self.f, fcntl.LOCK_EX)
+            except IOError, e:
+                # we should be more precise about whats a locking
+                # error and whats a random-other error
                 raise LockError(e)
 
-
     class _fcntl_ReadLock(_fcntl_FileLock):
+
         def __init__(self, filename):
+            # standard IO errors get exposed directly.
+            self._open(filename, 'rb')
             try:
-                fcntl.lockf(self._open(filename, 'rb'), fcntl.LOCK_SH)
-            except Exception, e:
+                fcntl.lockf(self.f, fcntl.LOCK_SH)
+            except IOError, e:
+                # we should be more precise about whats a locking
+                # error and whats a random-other error
                 raise LockError(e)
 
     WriteLock = _fcntl_WriteLock
     ReadLock = _fcntl_ReadLock
 
+
 except ImportError:
     try:
         import win32con, win32file, pywintypes
 
 
-        #LOCK_SH = 0 # the default
-        #LOCK_EX = win32con.LOCKFILE_EXCLUSIVE_LOCK
-        #LOCK_NB = win32con.LOCKFILE_FAIL_IMMEDIATELY
+        LOCK_SH = 0 # the default
+        LOCK_EX = win32con.LOCKFILE_EXCLUSIVE_LOCK
+        LOCK_NB = win32con.LOCKFILE_FAIL_IMMEDIATELY
 
         class _w32c_FileLock(_base_Lock):
             def _lock(self, filename, openmode, lockmode):
@@ -139,12 +144,13 @@ except ImportError:
 
         class _w32c_ReadLock(_w32c_FileLock):
             def __init__(self, filename):
-                _w32c_FileLock._lock(self, filename, 'rb', 0)
+                _w32c_FileLock._lock(self, filename, 'rb',
+                                     LOCK_NB)
 
         class _w32c_WriteLock(_w32c_FileLock):
             def __init__(self, filename):
                 _w32c_FileLock._lock(self, filename, 'wb',
-                                     win32con.LOCKFILE_EXCLUSIVE_LOCK)
+                                     LOCK_EX + LOCK_NB)
 
 
 
@@ -244,10 +250,3 @@ except ImportError:
         except ImportError:
             raise NotImplementedError("please write a locking method "
                                       "for platform %r" % sys.platform)
-
-
-
-
-
-
-
