@@ -1,5 +1,3 @@
-#! /usr/bin/env python
-
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
 # the Free Software Foundation; either version 2 of the License, or
@@ -15,7 +13,7 @@
 # Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 
 
-from bzrlib.xml import ElementTree, SubElement, Element, Serializer
+from bzrlib.xml_serializer import ElementTree, SubElement, Element, Serializer
 from bzrlib.inventory import ROOT_ID, Inventory, InventoryEntry
 import bzrlib.inventory as inventory
 from bzrlib.revision import Revision        
@@ -116,7 +114,7 @@ class _Serializer_v4(Serializer):
         else:
             raise BzrError("unknown kind %r" % kind)
 
-        ## mutter("read inventoryentry: %r" % (elt.attrib))
+        ## mutter("read inventoryentry: %r", elt.attrib)
 
         return ie
 
